/* 
   lmctl.c - Control Logitech USB Mice

   Copyright (C) 2004 Alexios Chouchoulas

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  

$Id: lmctl.c,v 2.1 2004/09/08 07:59:55 alexios Exp $

$Log: lmctl.c,v $
Revision 2.1  2004/09/08 07:59:55  alexios
Added MX1000 Laser support, courtesy of Markus Wiesner
<m_wiesner@gmx.net>.

Revision 2.0  2004/09/07 22:29:17  alexios
Stepped version to recover CVS repository after near-catastrophic disk
crash.

Revision 1.1.1.1  2004/09/07 22:28:18  alexios
Re-initialised to recover CVS repository after near-catastrophic disk crash.

Revision 1.1.1.1  2004/01/19 13:34:49  alexios
Initial revision.


*/

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <stropts.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/select.h>
#include <sys/time.h>
#include <pty.h>
#include <iconv.h>
#include "cmdline.h"
#include "lmctl.h"

#define __KERNEL__
#include <usb.h>

//#define DEBUG


#define VENDOR_LOGITECH 0x046d
#define TIMEOUT         0x1000


#define P6  buf [0]
#define P0  buf [1]
#define P4  buf [2]
#define P5  buf [3]
#define P8  buf [4]
#define P9  buf [5]
#define PB0 buf [6]
#define PB1 buf [7]



#define max(a,b) ((a) > (b)? (a): (b))


static char const rcsid[] = "$Id: lmctl.c,v 2.1 2004/09/08 07:59:55 alexios Exp $";


char *progname;


typedef struct {
	int      pid;
	char *   name;
	char *   model;

	int      has_csr;
	int      has_res;
	int      has_sms;

	int      is_dual;
} mouse_t;


mouse_t mice [] = {
	{0xc00e, "Wheel Mouse Optical",                    "M-BJ58",      0, 1, 0, 0},
	{0xc00f, "MouseMan Traveler",                      "M-BJ79",      0, 1, 0, 0},
	{0xc012, "MouseMan Dual Optical",                  "M-BL63B",     0, 1, 0, 0},
	{0xc01b, "MX310 Optical Mouse",                    "M-BP86",      0, 1, 0, 0},
	{0xc01d, "MX510 Optical Mouse",                    "M-BS81A",     0, 1, 1, 0},
	{0xc01e, "MX518 Optical Mouse",                    "M-BS81A",     0, 1, 0, 0},
	{0xc024, "MX300 Optical Mouse",                    "M-BP82",      0, 1, 0, 0},
	{0xc025, "MX500 Optical Mouse",                    "M-BP81A",     0, 1, 1, 0},
	{0xc031, "iFeel Mouse (silver)",                   "M-UT58A",     0, 1, 0, 0},
	{0xc501, "Mouse Receiver",                         "C-BA4-MSE",   1, 0, 0, 0},
	{0xc502, "Dual Receiver",                          "C-UA3-DUAL",  1, 0, 0, 1},
	{0xc504, "Receiver for Cordless Freedom Optical",  "C-BD9-DUAL",  1, 0, 0, 1},
	{0xc505, "Receiver for Cordless Elite Duo",        "C-BG17-DUAL", 1, 0, 0, 1},
	{0xc506, "Receiver for MX700 Optical Mouse",       "C-BF16-MSE",  1, 0, 1, 0},
	{0xc508, "Receiver for Cordless Optical TrackMan", "C-BA4-MSE",   1, 0, 1, 0},
	{0xc50b, "Receiver for Cordless MX Duo",           "967300-0403", 1, 0, 1, 0},
	/* From Markus Wiesner <m_wiesner@gmx.net> */
	{0xc50e, "Receiver for MX1000 Laser",              "M-RAG97/C-BN34", 1, 0, 1, 0},
	{0xc702, "Receiver for Cordless Presenter",        "C-UF15",      1, 0, 0, 0},

	{0x0000, NULL, NULL, 0, 0, 0}
};



static void
query (mouse_t * m, struct usb_device * device)
{
	usb_dev_handle * handle;
	unsigned char buf [80];
	unsigned int  addr;

	/* For dual receivers C502, C504 and C505, the mouse is the
	 * second device and uses an addr of 1 in the value and index
	 * fields' high byte. */

	addr = m->is_dual? 1<<8: 0;

	/* Try to open the device */

	handle = usb_open (device);

	if (handle == NULL) {
		perror ("Error opening USB device");
		exit (1);
	}

	/* Parse CSR Status (this one's huge) */

	if (m->has_csr) {
		if (usb_control_msg (handle, 0xc0, 0x09, 0x03|addr, 0x00|addr,
				     buf, 8, TIMEOUT) != 8) {
			perror("Writing to USB device");
			exit(1);
		}

		/* We have not obtained blocks P6 P0 P4 P5 P8 P9 PB0 PB1 */

		/* Is a C504 receiver busy? */
		
		if ((P0 == 0x3b) && (P4 == 0)) {
			printf("\t*** C504 receiver busy, try again later.\n");
		}
		
		/* Decode P0 */
		
		printf("\tReceiver type: ");
		switch (P0) {
		case 0x0f: printf ("C50E\n"); break;
		case 0x38: printf ("C501\n"); break;
		case 0x39: printf ("C502\n"); break;
		case 0x3b: printf ("C504\n"); break;
		case 0x3c: printf ("C508\n"); break;
		case 0x3d: printf ("C506\n"); break;
		case 0x3e: printf ("C505\n"); break;
		default: printf ("Unknown (type %x)\n", P0);
		}

		/* Decode P1 */

		printf("\tMouse type: ");
		switch (P4) {
		case 0x00: printf ("None\n"); break;
		case 0x04: printf ("Cordless Mouse\n"); break;
		case 0x05: printf ("Cordless Wheel Mouse\n"); break;
		case 0x06: printf ("Cordless MouseMan Wheel\n"); break;
		case 0x07: printf ("Cordless Wheel Mouse\n"); break;
		case 0x08: printf ("Cordless Wheel Mouse\n"); break;
		case 0x09: printf ("Cordless TrackMan Wheel\n"); break;
		case 0x0A: printf ("Trackman Live\n"); break;
		case 0x0C: printf ("Cordless TrackMan FX\n"); break;
		case 0x0D: printf ("Cordless MouseMan Optical\n"); break;
		case 0x0E: printf ("Cordless Optical Mouse\n"); break;
		case 0x0F: printf ("Cordless Mouse\n"); break;
		case 0x12: printf ("Cordless MouseMan Optical (2ch)\n"); break;
		case 0x13: printf ("Cordless Optical Mouse (2ch)\n"); break;
		case 0x14: printf ("Cordless Mouse (2ch)\n"); break;
		case 0x15: printf ("MX1000 Laser Cordless Mouse\n"); break;
		case 0x82: printf ("Cordless Optical TrackMan\n"); break;
		case 0x8A: printf ("Mx700 Cordless Optical Mouse\n"); break;
		case 0x8B: printf ("Mx700 Cordless Optical Mouse (2ch)\n"); break;
		default: printf ("Unknown (type %x)\n", P4);
		}

		/* Decode P5 */

		printf ("\tConnect button on mouse pressed: %s\n",
			P5 & 0x40? "yes": "no");
		printf ("\tActive RF Channel: %d\n",
			P5 & 0x08? 2: 1);
		printf ("\tBattery status: %d ", P5 & 0x07);
		switch (P5 & 0x07) {
		case 0: printf ("(unavailable)\n"); break;
		case 1:
		case 2: printf ("(critical)\n"); break;
		case 3:
		case 4: printf ("(low)\n"); break;
		case 5:
		case 6: printf ("(good)\n"); break;
		case 7: printf ("(full)\n"); break;
		}

		/* Decode P8 and P9 */

		printf ("\tCordless security ID: 0x%04x\n", P8 | (P9 << 8));

		/* Decode P6 */

		if (P6 & 0x20) {
			printf ("\tReceiver talking: yes\n");
			printf ("\tLock Request (CONNECT button pressed): %s\n",
				P6 & 0x80? "yes": "no");
			printf ("\tMouse Powerup (First time out of the box auto-locking): %s\n",
				P6 & 0x40? "yes": "no");
			printf ("\tReceiver Unlocked (Mouse disconnected): %s\n",
				P6 & 0x10? "yes": "no");
			printf ("\tWait Lock (Searching for mouse): %s\n",
				P6 & 0x08? "yes": "no");
		} else {
			printf ("\tReceiver talking: no\n");
		}
		
		/* Decode PB0 */

		if (PB0 < 0x80) 
			printf ("\tDevice physical shape type: %d\n", PB0);

		/* Decode PB1 */
		
		if (PB1 & 0x80) {
			printf ("\tDevice has two channels: %s\n",
				PB1 & 0x40? "yes": "no");
			printf ("\tDevice supports 800 cpi: %s\n",
				PB1 & 0x20? "yes": "no");
			printf ("\tDevice has horizontal roller: %s\n",
				PB1 & 0x10? "yes": "no");
			printf ("\tDevice has vertical roller: %s\n",
				PB1 & 0x08? "yes": "no");
			if ((PB1 & 7) < 7) {
				printf ("\tNumber of buttons on device: %d",
					(PB1 & 7) + 2);
			} else {
				printf ("\tNumber of buttons on device: 9 or more\n");
			}
		}
	}


	/* The RES extension */
	
	if (m->has_res) {
		if (usb_control_msg (handle, 0xc0, 0x01, 0x0e|addr, 0x00|addr,
				     buf, 1, TIMEOUT) != 1) {
			perror("Writing to USB device");
			exit(1);
		}
		
		assert ((buf [0] >= 3) || (buf [0] <= 6));

		printf ("\tResolution (RES): %d cpi\n", (buf [0] - 2) * 400);
	}
	

	/* The SMS extension */
	
	if (m->has_sms) {
		if (usb_control_msg (handle, 0xc0, 0x01, 0x17|addr, 0x00|addr,
				     buf, 1, TIMEOUT) != 1) {
			perror("Writing to USB device");
			exit(1);
		}
		
		printf ("\tSmartScroll (SMS): %s\n", buf [0]? "on": "off");
	}
	
	printf ("\n");
}


static void
configure_csr (mouse_t * m, struct usb_dev_handle * handle, int addr)
{
	unsigned char buf [80];
	
	/* Set the channel? */
	
	if (set_channel) {
		assert ((set_channel == 1) || (set_channel == 2));
		if (usb_control_msg (handle, 0x40, 0x02, 0x08|addr,
				     (set_channel - 1) | addr,
				     buf, 0, TIMEOUT) != 0) {
			perror("Writing to USB device");
			exit(1);
		}
		
		printf ("\tRF Channel set to %d\n", set_channel);
	}
	
	/* Unlock? */
	
	if (set_unlock) {
		if (usb_control_msg (handle, 0x40, 0x02, 0x06|addr, 0x1|addr,
				     buf, 0, TIMEOUT) != 0) {
			perror("Writing to USB device");
			exit(1);
		}
		
		printf ("\tReceiver unlocked, searching for a mouse.\n");
	}
	
	/* Lock? */
	
	if (set_lock) {
		if (usb_control_msg (handle, 0x40, 0x02, 0x06|addr, 0x0|addr,
				     buf, 0, TIMEOUT) != 0) {
			perror("Writing to USB device");
			exit(1);
		}
		
		printf ("\tReceiver locked, no longer searching for a mouse.\n");
	}
	
	/* Clear? */

	if (set_clear) {
		if (usb_control_msg (handle, 0x40, 0x09, 0x04|addr, 0x0|addr,
				     buf, 0, TIMEOUT) != 0) {
			perror("Writing to USB device");
			exit(1);
		}
		
		printf ("\tReceiver cleared, no longer associated with a mouse.\n");
	}
}


static void
configure_res (mouse_t * m, struct usb_dev_handle * handle)
{
	unsigned char buf [80];
	if (!set_res) return;

	assert ((set_res == 400) || (set_res == 800));
	if (usb_control_msg (handle, 0x40, 0x02, 0x000e, (set_res / 400) + 2,
			     buf, 0, TIMEOUT) != 0) {
		perror("Writing to USB device");
		exit(1);
		}
	
	printf ("\tResolution set to %d cpi\n", set_res);
}


static void
configure_sms (mouse_t * m, struct usb_dev_handle * handle)
{
	unsigned char buf [80];
	if (!set_sms) return;
	assert ((set_sms == 1) || (set_sms == -1));
	if (usb_control_msg (handle, 0x40, 0x02, 0x0017,
			     set_sms == 1? 0x0001: 0x0000,
			     buf, 0, TIMEOUT) != 0) {
		perror("Writing to USB device");
		exit(1);
	}
	
	if (set_sms == 1) printf ("\tSmartScroll enabled\n");
	else printf ("\tSmartScroll disabled\n");
}




static void
configure (mouse_t * m, struct usb_device * device)
{
	usb_dev_handle * handle;
	unsigned int  addr;

	/* For dual receivers C502, C504 and C505, the mouse is the
	 * second device and uses an addr of 1 in the value and index
	 * fields' high byte. */

	addr = m->is_dual? 1<<8: 0;

	/* Try to open the device */

	handle = usb_open (device);

	if (handle == NULL) {
		perror ("Error opening USB device");
		exit (1);
	}

	if (m->has_csr) configure_csr (m, handle, addr);
	if (m->has_res) configure_res (m, handle);
	if (m->has_sms) configure_sms (m, handle);

	printf ("\n");
}


static mouse_t *
find_mouse (struct usb_device * device)
{
	int i;

	for (i = 0; mice [i].pid; i++) {
		if (device->descriptor.idProduct == mice [i].pid) {
			return &mice [i];
		}
	}

	return NULL;
}



int
scan_bus( struct usb_bus* bus )
{
	struct usb_device* roottree = bus->devices;
	struct usb_device *device;
	mouse_t * m;
	
	for (device = roottree; device; device = device->next) {

		/* Matching */

		if (match_pid && (device->descriptor.idProduct != match_pid)) continue;
		if ((match_dev != NULL) &&
		    (strcmp (device->filename, match_dev))) continue;
		if ((match_bus != NULL) &&
		    (strcmp (device->bus->dirname, match_bus))) continue;

		/* Is this a Logitech device? */

		if (device->descriptor.idVendor != VENDOR_LOGITECH) {
			if (command == cmd_scan) {
				printf ("%s.%s: %04x:%04x Not a Logitech device\n",
					device->bus->dirname,
					device->filename,
					device->descriptor.idVendor,
					device->descriptor.idProduct);
			}
			continue;
		}

		/* Do we support this device? If so, list it. */

		if ((m = find_mouse (device)) != NULL) {
			
			printf ("%s.%s: %04x:%04x %s (%s) Caps: %s%s%s\n",
				device->bus->dirname,
				device->filename,
				device->descriptor.idVendor,
				device->descriptor.idProduct,
				m->name,
				m->model,
				m->has_csr? "CSR ": "",
				m->has_res? "RES ": "",
				m->has_sms? "SMS ": ""
				);

		} else {
			printf ("%s.%s: %04x:%04x Unknown or Unsupported Logitech device\n",
				device->bus->dirname,
				device->filename,
					device->descriptor.idVendor,
				device->descriptor.idProduct);
			continue;
		}


		/* Run commands */

		if (command == cmd_inquire) query (m, device);
		if (command == cmd_set) configure (m, device);

	}
	
	return 1;
}

void
run ()
{
	struct usb_bus* bus;
	
	usb_init();
	usb_find_busses();
	usb_find_devices();
	
	for (bus = usb_busses; bus != NULL; bus = bus->next) {
		if (!scan_bus(bus)) {
			printf ("Exiting happily.\n");
			exit (EXIT_SUCCESS);
		}
	}
}


int
main (int argc, char **argv)
{
	progname = argv[0];
	
	cmdline_parse (argc, argv);

	run();
	return EXIT_SUCCESS;
}
